
// Operations on block distributed ranges.


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRangeRep': Block-distributed template range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockRangeRep : public RangeRep {
public :
  BlockRangeRep(const int _extent, RangeRep* _kernel, const int _blockSize) :
          RangeRep(_extent, _kernel, _kernel->lev() + 1, _kernel->dim(),
                   DIST_BLOCK),
          blockSize(_blockSize) {
    kernel->incr() ;

    ker_hi = (extent - 1) / blockSize ;
  }

  ~BlockRangeRep() {
    kernel->decr() ;
  }

  RangeRep* shell() const ;

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  void location(Location* loc, const int glb) const ;

  void block(const int ker_glb, Block* blk) const ;

  RangeRep* subker() const ;

  int volume() const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;
protected :
  int blockSize ;

  int ker_hi ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockSubRangeRep': Block-distributed subrange.               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockSubRangeRep : public SubRangeRep {
public :
  BlockSubRangeRep(const int _extent,
		   const int _base, const int _stride,
		   RangeRep* _parent,
		   RangeRep* _kernel, const int _blockSize) :
          SubRangeRep(_extent, _base, _stride, _parent,
                      _kernel, _kernel->lev() + 1, _kernel->dim(),
                      DIST_BLOCK),
          blockSize(_blockSize) {

    if(stride > 0) {
      ker_lo = base / blockSize ;
      ker_hi = (base + stride * (extent - 1)) / blockSize ;
    }
    else {
      ker_hi = base / blockSize ;
      ker_lo = (base + stride * (extent - 1)) / blockSize ;
    }
  }

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  void location(Location* loc, const int glb) const ;

  RangeRep* shell() const ;

  void block(const int ker_glb, Block* blk) const ;

  RangeRep* subker() const ;

  int volume() const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;
protected :
  int blockSize ;

  int ker_lo, ker_hi ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedBlockRangeRep', `ExtendedBlockSubRangeRep':        //
//     Block-distributed range array with ghost areas.                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ExtendedBlockRangeRep : public BlockRangeRep {
public :
  ExtendedBlockRangeRep(const int _extent,
                        RangeRep* _kernel, const int _blockSize,
                        const int _wlo, const int _whi) :
          BlockRangeRep(_extent, _kernel, _blockSize), wlo(_wlo), whi(_whi) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  RangeRep* shell() const ;

  int disp(const int sub) const ;

  int offset(const Location* i) const ;

  int volume() const ;

protected :
  int wlo, whi ;
  
MEMORY_POOL_DECS
} ;


class ExtendedBlockSubRangeRep : public BlockSubRangeRep {
public :
  ExtendedBlockSubRangeRep(const int _extent,
                           const int _base, const int _stride,
			   RangeRep* _parent,
			   RangeRep* _kernel, const int _blockSize,
                           const int _wlo, const int _whi) :
    BlockSubRangeRep(_extent, _base, _stride, _parent, _kernel, _blockSize),
    wlo(_wlo), whi(_whi) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  RangeRep* shell() const ;

  int disp(const int sub) const ;

  int offset(const Location* i) const ;

  int volume() const ;

protected :
  int wlo, whi ;
  
MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PackedBlockSubRangeRep':                                     //
//      Block-distributed range of array with ``packed'' layout.       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class PackedBlockSubRangeRep : public BlockSubRangeRep {
public :
  PackedBlockSubRangeRep(const int _extent,
                         const int _base, const int _stride,
                         RangeRep* _parent,
                         RangeRep* _kernel, const int _blockSize,
                         const int _packing) :
        BlockSubRangeRep(_extent, _base, _stride,
                         _parent, _kernel, _blockSize),
        packing(_packing), width((_blockSize + _packing - 1) / _packing) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  RangeRep* shell() const ;

  int disp(const int sub) const ;
  int step(const int sub_stp) const ;

  int offset(const Location* i) const ;

  int volume() const ;

protected :
  int packing ;
  int width ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `BlockCyclicRangeRep', `BlockCyclicSubRangeRep':            //
//     Block-cyclically distributed ranges.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockCyclicRangeRep : public BlockRangeRep {
public :
  BlockCyclicRangeRep(const int _extent, RangeRep* _kernel,
                      const int _blockSize) :
          BlockRangeRep(_extent, _kernel, _blockSize) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  void location(Location* loc, const int glb) const ;

  int offset(const Location* i) const ;

  int volume() const ;
} ;


class BlockCyclicSubRangeRep : public BlockSubRangeRep {
public :
  BlockCyclicSubRangeRep(const int _extent,
		         const int _base, const int _stride,
		         RangeRep* _parent,
		         RangeRep* _kernel, const int _blockSize) :
          BlockSubRangeRep(_extent, _base, _stride, _parent,
                           _kernel, _blockSize) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  void location(Location* loc, const int glb) const ;

  int offset(const Location* i) const ;

  int volume() const ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PackedBlockCyclicSubRangeRep':                               //
//      Block-cyclically distributed range of array with ``packed''    //
//      layout.                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class PackedBlockCyclicSubRangeRep : public PackedBlockSubRangeRep {
public :
  PackedBlockCyclicSubRangeRep(const int _extent,
			       const int _base, const int _stride,
			       RangeRep* _parent,
			       RangeRep* _kernel, const int _blockSize,
			       const int _packing) :
          PackedBlockSubRangeRep(_extent, _base, _stride, _parent,
		                 _kernel, _blockSize, _packing) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  void location(Location* loc, const int glb) const ;

  int offset(const Location* i) const ;

  int volume() const ;
} ;


// -------------------------------------------------------------
// Things to do
//
//   Simplified sub-range classes for contiguous subranges.
//
//   Extended block-cyclic subclasses.
//
//   `show' methods for subclasses.

