
// Range representation base class.


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Pool' : Modified from Stroustrup.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Pool {
  struct Link {Link* next ; } ;

  unsigned esize ;
  Link* head ;

  void grow() ;

  int nalloc ;  // debug
public :
  void init(unsigned n) ;

  void* alloc() ;
  void free(void* b) ;
} ;

inline void* Pool :: alloc() {
  if(head == 0) grow() ;
  Link* p = head ;
  head = p->next ;
  nalloc++ ;  // debug
  return p ;
}

inline void Pool :: free(void* b) {
  Link* p = (Link*) b ;
  p->next = head ;
  head = p ;
  nalloc-- ;  // debug
}


#define POOL_NEW void* operator new(size_t) { return pool.alloc() ; }

#define POOL_DELETE void operator delete(void* p) { pool.free(p) ; }

#define POOL_STATE static Pool pool ; friend void RangeInit() ;

#define MEMORY_POOL_DEFN(T) Pool T :: pool ;

#define MEMORY_POOL_DECS public : POOL_NEW POOL_DELETE protected : POOL_STATE

void RangeInit() ;


enum Format {DIST_PRIMITIVE, DIST_COLLAPSED, DIST_BLOCK, DIST_CYCLIC,
             DIST_STEP, DIST_IRREG} ;

#define MAX_LEVELS 2


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// struct `Location': The local subscript.                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //


struct Location : public Coord {

  int tem ;
  int sub ;
  int blk ;

  // Cosmetic support for ghost regions:

  Location operator<<(const int shift) {
    Location shifted(*this) ;
    shifted.sub -= shift ;
    return shifted ;
  }

  Location operator>>(const int shift) {
    Location shifted(*this) ;
    shifted.sub += shift ;
    return shifted ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// struct `Block'                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Block {
  int count ;

  int glb_bas ;
  int glb_stp ;

  int tem_bas ;
  int tem_stp ;

  int sub_bas ;
  int sub_stp ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `RangeRep': Base class for representation of ranges.          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class StrideRep ;

class RangeRep {
public :
  // Constructors and destructors

  RangeRep(const int _extent,
           RangeRep* _kernel, const int _level, Dimension _dimension,
           const Format _distFormat) : 
      extent(_extent), kernel(_kernel), level(_level), dimension(_dimension),
      distFormat(_distFormat),
      refCount(0) {}

  virtual ~RangeRep() {}

  // Subranges.

  virtual RangeRep* subrng(const int _extent,
                           const int _base, const int _stride) = 0 ;

  // Inquiries.

  int size() const {return extent ; }

  Dimension dim() const {return dimension ; }

  Format format() {return distFormat ; }

  virtual RangeRep* tem() {return this ; }
  virtual int bas()  const {return 0 ; }
  virtual int str()  const {return 1 ; }

  int lev() const {return level ; }
  virtual RangeRep* ker() const {return kernel ; }

  virtual RangeRep* shell() const {
    AdlibError("RangeRep :: shell: member undefined on this range") ;
    return 0 ;
  }

  virtual void location(Location* loc, const int glb) const = 0 ;

  virtual void block(const int, Block*) const {
    AdlibError("RangeRep :: block: member undefined on this range") ;
  }

  virtual int local(int*) const {
    AdlibError("RangeRep :: local: member undefined on this range") ;
    return 0 ;
  }

  virtual RangeRep* subker() const {
    return kernel ;
  }

  // Strides

  virtual int volume() const = 0 ;

  virtual int offset(const Location* i) const {
    return i->sub ;
  }

  virtual int disp(const int sub) const {
    return sub ;
  }

  virtual int step(const int sub_stp) const {
    return sub_stp ;
  }

  // idx

  virtual int idx(const Location* loc) const {
    return (loc->tem - bas()) / str() ;
  }

  // Reference counting.

  void incr() {refCount++ ; }
  void decr() {if((--refCount) == 0) delete this ; }
  
  // Interim support for PCRC Fortran interface:

  virtual int glb(const int sub, const int) const {return sub ; }

  virtual void show() = 0 ;
protected :
  int extent ;
  RangeRep* kernel ;
  int level ;
  Dimension dimension ;
  const Format distFormat ;

private :
  int refCount ;

  // Interim support for PCRC Fortran interface:

  int g_lb, g_stride ;

friend class Range ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `SubRangeRep': abstract subrange.                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class SubRangeRep : public RangeRep {
public :
  SubRangeRep(const int _extent, const int _base, const int _stride,
              RangeRep* _parent,
              RangeRep* _kernel, const int _level, Dimension _dimension,
              const Format _distFormat) :
          RangeRep(_extent, _kernel, _level, _dimension, _distFormat),
          base(_base), stride(_stride), parent(_parent) {
    parent->incr() ;
  }

  ~SubRangeRep() {
    parent->decr() ;
  }

  RangeRep* tem() {return parent ; }

  int str() const {return stride ; }
  int bas() const {return base ; }

protected :

  int base, stride ;
  RangeRep* parent ;
} ;


// ---------------------------------------------------
//
// Things to do:
//
//   Things that return `RangeRep*' should return `const RangeRep*'?

