
// Operations on block distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"

inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRangeRep': Block-distributed template range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* BlockRangeRep :: subrng(const int _extent,
                                  const int _base, const int _stride) {
  return new BlockSubRangeRep(_extent, _base, _stride, this,
                              kernel, blockSize) ;
}

RangeRep* BlockRangeRep :: shell() const {
  return new CollapsedRangeRep(blockSize) ;
}

void BlockRangeRep :: location(Location* loc, const int glb) const {

  loc->tem = glb ;
  loc->sub = glb % blockSize ;
  loc->crd = glb / blockSize ;
  loc->dim = dimension ;
}

void BlockRangeRep :: block(const int ker_glb, Block* blk) const {

  const int blockStart  = ker_glb * blockSize ;

  blk->sub_stp = 1 ;
  blk->tem_stp = 1 ;
  blk->glb_stp = 1 ;

  if(ker_glb <= ker_hi) {
    if(blockStart + blockSize > extent)
      blk->count = extent - blockStart ;
    else
      blk->count = blockSize ;
  }
  else
    blk->count = 0 ;

  blk->sub_bas = 0 ;
  blk->tem_bas = blockStart ;
  blk->glb_bas = blockStart ;
}

RangeRep* BlockRangeRep :: subker() const {

  return kernel->subrng(ker_hi + 1, 0, 1) ;
}

int BlockRangeRep :: volume() const {return blockSize ; }

int BlockRangeRep :: glb(const int sub, const int ker_glb) const {
  return ker_glb * blockSize + sub ;
}

void BlockRangeRep :: show() {
  AdlibPrintf("BlockRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ;
}

MEMORY_POOL_DEFN(BlockRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockSubRangeRep': block-distributed subrange.               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* BlockSubRangeRep :: subrng(const int _extent,
                                     const int _base, const int _stride) {
  return new BlockSubRangeRep(_extent,
			      base + stride * _base, stride * _stride,
			      parent, kernel, blockSize) ;
}

RangeRep* BlockSubRangeRep :: shell() const {
  return new CollapsedRangeRep(blockSize) ;
}

void BlockSubRangeRep :: location(Location* loc, const int glb) const {
  const int tem = base + stride * glb ;

  loc->tem = tem ;
  loc->sub = tem % blockSize ;
  loc->crd = tem / blockSize ;
  loc->dim = dimension ;
}

void BlockSubRangeRep :: block(const int ker_glb, Block* blk) const {

  blk->tem_stp = stride ;
  blk->sub_stp = stride ;
  blk->glb_stp = 1 ;

  if(ker_glb >= ker_lo && ker_glb <= ker_hi) {
    const int blockStart = ker_glb * blockSize ;

    const int blockStartOffset  = blockStart - base ;
    const int blockFinishOffset = blockStartOffset + blockSize - 1 ;

    int i_hi, i_lo ;
    if(stride > 0) {
      i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
      i_hi = min(blockFinishOffset / stride, extent - 1) ;
    }
    else {
      i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
      i_hi = min(blockStartOffset / stride, extent - 1) ;
    }
   
    blk->count = i_hi - i_lo + 1 ;

    blk->tem_bas = base + stride * i_lo ;
    blk->sub_bas = blk->tem_bas - blockStart ;
    blk->glb_bas = i_lo ;
  }
  else {
    blk->count   = 0 ;

    blk->tem_bas = 0 ;
    blk->sub_bas = 0 ;
    blk->glb_bas = 0 ;
  }
}

RangeRep* BlockSubRangeRep :: subker() const {

  int ker_bas, ker_str ;

  if(stride > 0) {
    ker_bas = ker_lo ;
    ker_str = 1 ;
  }
  else {
    ker_bas = ker_hi ;
    ker_str = -1 ;
  }

  return kernel->subrng(ker_hi - ker_lo + 1, ker_bas, ker_str) ;
}

int BlockSubRangeRep :: volume() const {return blockSize ; }

int BlockSubRangeRep :: glb(const int sub, const int ker_glb) const {
  return (ker_glb * blockSize + sub - base) / stride ;
}

void BlockSubRangeRep :: show() {
  AdlibPrintf("BlockSubRangeRep :: show: "
              "size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ;
}

MEMORY_POOL_DEFN(BlockSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedBlockRangeRep', `ExtendedBlockSubRangeRep':        //
//     Block-distributed range of array with ghost areas.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* ExtendedBlockRangeRep :: subrng(const int _extent,
                                          const int _base, const int _stride) {
  return new ExtendedBlockSubRangeRep(_extent, _base, _stride, this,
                                      kernel, blockSize, wlo, whi) ;
}

RangeRep* ExtendedBlockRangeRep :: shell() const {
  return new ExtendedCollapsedRangeRep(blockSize, wlo, whi) ;
}

int ExtendedBlockRangeRep :: disp(const int sub) const {
  return sub + wlo ;
}

int ExtendedBlockRangeRep :: offset(const Location* i) const {
  return i->sub + wlo ;
}

int ExtendedBlockRangeRep :: volume() const {
  return wlo + blockSize + whi ;
}

RangeRep* ExtendedBlockSubRangeRep :: subrng(const int _extent,
                                             const int _base,
                                             const int _stride) {
  return new ExtendedBlockSubRangeRep(_extent,
				      base + stride * _base, stride * _stride,
				      parent, kernel, blockSize, wlo, whi) ;
}

RangeRep* ExtendedBlockSubRangeRep :: shell() const {
  return new ExtendedCollapsedRangeRep(blockSize, wlo, whi) ;
}

int ExtendedBlockSubRangeRep :: disp(const int sub) const {
  return sub + wlo ;
}

int ExtendedBlockSubRangeRep :: offset(const Location* i) const {
  return i->sub + wlo ;
}

int ExtendedBlockSubRangeRep :: volume() const {
  return wlo + blockSize + whi ;
}


MEMORY_POOL_DEFN(ExtendedBlockRangeRep)

MEMORY_POOL_DEFN(ExtendedBlockSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `PackedBlockSubRangeRep':                                   //
//      Block-distributed ranges of array with ``packed'' layout.      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* PackedBlockSubRangeRep :: subrng(const int _extent,
                                           const int _base, const int _stride) {
  return new PackedBlockSubRangeRep(_extent,
				    base + stride * _base, stride * _stride,
				    parent, kernel, blockSize, packing) ;
}

RangeRep* PackedBlockSubRangeRep :: shell() const {
  AdlibError("PackedBlockSubRangeRep :: shell: "
             "member undefined on this range") ;
  return 0 ;
}

int PackedBlockSubRangeRep :: disp(const int sub) const {
  return sub / packing ;
}

int PackedBlockSubRangeRep :: step(const int sub_stp) const {
  return sub_stp / packing ;
}

int PackedBlockSubRangeRep :: offset(const Location* i) const {
  return i->sub / packing ;
}

int PackedBlockSubRangeRep :: volume() const {
  return width ;
}

MEMORY_POOL_DEFN(PackedBlockSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `BlockCyclicRangeRep', `BlockCyclicSubRangeRep':            //
//     Block-cyclically distributed ranges.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* BlockCyclicRangeRep :: subrng(const int _extent,
                                        const int _base, const int _stride) {
  return new BlockCyclicSubRangeRep(_extent, _base, _stride, this,
                                    kernel, blockSize) ;
}

void BlockCyclicRangeRep :: location(Location* loc, const int glb) const {
  const int numProcs = dimension.size() ;

  const int ker = glb / blockSize ;

  loc->tem = glb ;
  loc->sub = glb % blockSize ;
  loc->blk = ker / numProcs ;
  loc->crd = ker % numProcs ;
  loc->dim = dimension ;
}

int BlockCyclicRangeRep :: offset(const Location* i) const {
  return i->sub + blockSize * i->blk ;
}

int BlockCyclicRangeRep :: volume() const {
  return blockSize * kernel->volume() ;
}

RangeRep* BlockCyclicSubRangeRep :: subrng(const int _extent,
                                           const int _base, const int _stride) {
  return new BlockCyclicSubRangeRep(_extent,
			            base + stride * _base, stride * _stride,
			            parent, kernel, blockSize) ;
}

void BlockCyclicSubRangeRep :: location(Location* loc, const int glb) const {
  const int numProcs = dimension.size() ;

  const int tem = base + stride * glb ;
  const int ker = tem / blockSize ;

  loc->tem = tem ;
  loc->sub = tem % blockSize ;
  loc->blk = ker / numProcs ;
  loc->crd = ker % numProcs ;
  loc->dim = dimension ;
}

int BlockCyclicSubRangeRep :: offset(const Location* i) const {
  return i->sub + blockSize * i->blk ;
}

int BlockCyclicSubRangeRep :: volume() const {
  return blockSize * kernel->volume() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PackedBlockCyclicSubRangeRep':                               //
//      Block-cyclically distributed range of array with ``packed''    //
//      layout.                                                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* PackedBlockCyclicSubRangeRep :: subrng(const int _extent,
                                                 const int _base,
                                                 const int _stride) {
  return new PackedBlockCyclicSubRangeRep(_extent,
			                  base + stride * _base,
                                          stride * _stride,
			                  parent, kernel, blockSize, packing) ;
}

void PackedBlockCyclicSubRangeRep :: location(Location* loc,
                                              const int glb) const {
  const int numProcs = dimension.size() ;

  const int tem = base + stride * glb ;
  const int ker = tem / blockSize ;

  loc->tem = tem ;
  loc->sub = tem % blockSize ;
  loc->blk = ker / numProcs ;
  loc->crd = ker % numProcs ;
  loc->dim = dimension ;
}

int PackedBlockCyclicSubRangeRep :: offset(const Location* i) const {
  return i->sub / packing + width * i->blk ;
}

int PackedBlockCyclicSubRangeRep :: volume() const {
  return width * kernel->volume() ;
}


// -------------------------------------------------------------
// Things to do
//
//   `subker' returns non-empty subrange even for extent = 0
//
//   Simplified sub-range classes for contiguous subranges.
//
//   Extended block-cyclic subclasses.
//
//   `show' methods for subclasses.

