
// Operations on stepped block distributed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"

inline int max(int x, int y) {return x > y ? x : y ; }
inline int min(int x, int y) {return x < y ? x : y ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepRangeRep': stepped block-distributed template range.     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* StepRangeRep :: subrng(const int _extent,
                                 const int _base, const int _stride) {
  return new StepSubRangeRep(_extent, _base, _stride, this,
                             kernel, blockSize, stepCrd) ;
}

void StepRangeRep :: location(Location* loc, const int glb) const {

  loc->tem = glb ;
  if(glb < stepSub) {
    loc->sub = glb % (blockSize + 1) ;
    loc->crd = glb / (blockSize + 1) ;
  }
  else {
    loc->sub = (glb - stepCrd) % blockSize ;
    loc->crd = (glb - stepCrd) / blockSize ;
  }
  loc->dim = dimension ;
}

RangeRep* StepRangeRep :: shell() const {
  return new CollapsedRangeRep(blockSize) ;
}

void StepRangeRep :: block(const int ker_glb, Block* blk) const {

  // `ker_glb' is a global subscript in the kernel range.

  blk->tem_stp = 1 ;
  blk->sub_stp = 1 ;
  blk->glb_stp = 1 ;

  blk->sub_bas = 0 ;
  if(ker_glb < stepCrd) {
    blk->tem_bas = (blockSize + 1) * ker_glb ;
    blk->glb_bas = blk->tem_bas ;

    blk->count   = blockSize + 1 ;
  }
  else {
    blk->tem_bas = blockSize * ker_glb + stepCrd ;
    blk->glb_bas = blk->tem_bas ;

    blk->count   = blockSize ;
  }
}

int StepRangeRep :: volume() const {
  return stepCrd ? blockSize + 1 : blockSize ;
}

int StepRangeRep :: glb(const int sub, const int ker_glb) const {
  if(ker_glb < stepCrd)
    return (blockSize + 1) * ker_glb + sub ;
  else
    return blockSize * ker_glb + stepCrd + sub ;
}

void StepRangeRep :: show() {
  AdlibPrintf("StepRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ; 
}

MEMORY_POOL_DEFN(StepRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepSubRangeRep': stepped block-distributed subrange.        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* StepSubRangeRep :: subrng(const int _extent,
                                    const int _base, const int _stride) {
  return new StepSubRangeRep(_extent,
			     base + stride * _base, stride * _stride,
			     parent, kernel, blockSize, stepCrd) ;
}

void StepSubRangeRep :: location(Location* loc, const int glb) const {
  const int tem = base + stride * glb ;

  loc->tem = tem ;
  if(glb < stepSub) {
    loc->sub = tem % (blockSize + 1) ;
    loc->crd = tem / (blockSize + 1) ;
  }
  else {
    loc->sub = (tem - stepCrd) % blockSize ;
    loc->crd = (tem - stepCrd) / blockSize ;
  }
  loc->dim = dimension ;
}

RangeRep* StepSubRangeRep :: shell() const {
  return new CollapsedRangeRep(blockSize) ;
}

void StepSubRangeRep :: block(const int ker_glb, Block* blk) const {

  blk->tem_stp = stride ;
  blk->sub_stp = stride ;
  blk->glb_stp = 1 ;

  if(ker_glb >= ker_lo && ker_glb <= ker_hi) {
    int blockStart, blockStartOffset, blockFinishOffset ;
    if(ker_glb < stepCrd) {
      blockStart        = ker_glb * (blockSize + 1) ;
      blockStartOffset  = blockStart - base ;
      blockFinishOffset = blockStartOffset + blockSize ;
    }
    else {
      blockStart        = ker_glb * blockSize + stepCrd;
      blockStartOffset  = blockStart - base ;
      blockFinishOffset = blockStartOffset + blockSize - 1 ;
    }

    int i_lo, i_hi ;

    if(stride > 0) {
      i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
      i_hi = min(blockFinishOffset / stride, extent - 1) ;
    }
    else {
      i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
      i_hi = min(blockStartOffset / stride, extent - 1) ;
    }

    blk->tem_bas = base + stride * i_lo ;
    blk->sub_bas = blk->tem_bas - blockStart ;
    blk->glb_bas = i_lo ;

    blk->count   = i_hi - i_lo + 1 ;
  }
  else {
    blk->tem_bas = 0 ;
    blk->sub_bas = 0 ;
    blk->glb_bas = 0 ;

    blk->count   = 0 ;
  }
}

RangeRep* StepSubRangeRep :: subker() const {

  int ker_bas, ker_str ;
  if(stride > 0) {
    ker_bas = ker_lo ;
    ker_str = 1 ;
  }
  else {
    ker_bas = ker_hi ;
    ker_str = -1 ;
  }

  return kernel->subrng(ker_hi - ker_lo + 1, ker_bas, ker_str) ;
}

int StepSubRangeRep :: glb(const int sub, const int ker_glb) const {
  if(ker_glb < stepCrd)
    return ((blockSize + 1) * ker_glb + sub - base) / stride ;
  else
    return (blockSize * ker_glb + stepCrd + sub - base) / stride ;
}

int StepSubRangeRep :: volume() const {
  return stepCrd ? blockSize + 1 : blockSize ;
}

void StepSubRangeRep :: show() {
  AdlibPrintf("StepSubRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ; 
}

MEMORY_POOL_DEFN(StepSubRangeRep)


// -------------------------------------------------------------
// Things to do
//
//   `subker' returns non-empty subrange even for extent = 0
//        (not a problem?)

